@@ -6,6 +6,7 @@
 #include <sys/uio.h>
 #if defined(__FreeBSD__)
 #include <sys/endian.h>
+#include <sys/capsicum.h>
 #elif defined(__APPLE__)
 #include <machine/endian.h>
 #include <libkern/OSByteOrder.h>
@@ -69,6 +70,10 @@
 #define unveil(s, p) (0)
 #endif
 
+#ifndef __FreeBSD__
+#define cap_enter() (0)
+#endif
+
 #ifndef HAVE_LINUX_LANDLOCK_H
 #define landlock_no_fs() (0)
 #else
