@@ -19,7 +19,6 @@
 #include <sys/uio.h>
 #include <sys/socket.h>
 #include <sys/wait.h>
-#include <sys/resource.h>
 #include <sys/mman.h>
 
 #include <errno.h>
@@ -244,80 +243,6 @@ request_packed_object_raw(uint8_t **outbuf, off_t *size, size_t *hdrlen,
 	return NULL;
 }
 
-static void
-set_max_datasize(void)
-{
-	struct rlimit rl;
-
-	if (getrlimit(RLIMIT_DATA, &rl) != 0)
-		return;
-
-	rl.rlim_cur = rl.rlim_max;
-	setrlimit(RLIMIT_DATA, &rl);
-}
-
-static const struct got_error *
-start_pack_privsep_child(struct got_pack *pack, struct got_packidx *packidx)
-{
-	const struct got_error *err = NULL;
-	int imsg_fds[2];
-	pid_t pid;
-	struct imsgbuf *ibuf;
-
-	ibuf = calloc(1, sizeof(*ibuf));
-	if (ibuf == NULL)
-		return got_error_from_errno("calloc");
-
-	pack->privsep_child = calloc(1, sizeof(*pack->privsep_child));
-	if (pack->privsep_child == NULL) {
-		err = got_error_from_errno("calloc");
-		free(ibuf);
-		return err;
-	}
-	pack->child_has_tempfiles = 0;
-	pack->child_has_delta_outfd = 0;
-
-	if (socketpair(AF_UNIX, SOCK_STREAM, PF_UNSPEC, imsg_fds) == -1) {
-		err = got_error_from_errno("socketpair");
-		goto done;
-	}
-
-	pid = fork();
-	if (pid == -1) {
-		err = got_error_from_errno("fork");
-		goto done;
-	} else if (pid == 0) {
-		set_max_datasize();
-		got_privsep_exec_child(imsg_fds, GOT_PATH_PROG_READ_PACK,
-		    pack->path_packfile);
-		/* not reached */
-	}
-
-	if (close(imsg_fds[1]) == -1)
-		return got_error_from_errno("close");
-	pack->privsep_child->imsg_fd = imsg_fds[0];
-	pack->privsep_child->pid = pid;
-	imsg_init(ibuf, imsg_fds[0]);
-	pack->privsep_child->ibuf = ibuf;
-
-	err = got_privsep_init_pack_child(ibuf, pack, packidx);
-	if (err) {
-		const struct got_error *child_err;
-		err = got_privsep_send_stop(pack->privsep_child->imsg_fd);
-		child_err = got_privsep_wait_for_child(
-		    pack->privsep_child->pid);
-		if (child_err && err == NULL)
-			err = child_err;
-	}
-done:
-	if (err) {
-		free(ibuf);
-		free(pack->privsep_child);
-		pack->privsep_child = NULL;
-	}
-	return err;
-}
-
 static const struct got_error *
 read_packed_object_privsep(struct got_object **obj,
     struct got_repository *repo, struct got_pack *pack,
@@ -326,7 +251,7 @@ read_packed_object_privsep(struct got_object **obj,
 	const struct got_error *err = NULL;
 
 	if (pack->privsep_child == NULL) {
-		err = start_pack_privsep_child(pack, packidx);
+		err = got_pack_start_privsep_child(pack, packidx);
 		if (err)
 			return err;
 	}
@@ -342,7 +267,7 @@ read_packed_object_raw_privsep(uint8_t **outbuf, off_t *size, size_t *hdrlen,
 	const struct got_error *err = NULL;
 
 	if (pack->privsep_child == NULL) {
-		err = start_pack_privsep_child(pack, packidx);
+		err = got_pack_start_privsep_child(pack, packidx);
 		if (err)
 			return err;
 	}
@@ -425,7 +350,7 @@ got_object_read_raw_delta(uint64_t *base_size, uint64_t *result_size,
 	}
 
 	if (pack->privsep_child == NULL) {
-		err = start_pack_privsep_child(pack, packidx);
+		err = got_pack_start_privsep_child(pack, packidx);
 		if (err)
 			return err;
 	}
@@ -452,55 +377,6 @@ got_object_read_raw_delta(uint64_t *base_size, uint64_t *result_size,
 	    pack->privsep_child->ibuf);
 }
 
-/*
- * XXX This function does not really belong in object.c. It is only here
- * because it needs start_pack_privsep_child(); relevant code should
- * probably be moved to pack.c/pack_create.c.
- */
-const struct got_error *
-got_object_prepare_delta_reuse(struct got_pack **pack,
-    struct got_packidx *packidx, int delta_outfd, struct got_repository *repo)
-{
-	const struct got_error *err = NULL;
-	char *path_packfile = NULL;
-
-	err = got_packidx_get_packfile_path(&path_packfile,
-	    packidx->path_packidx);
-	if (err)
-		return err;
-
-	*pack = got_repo_get_cached_pack(repo, path_packfile);
-	if (*pack == NULL) {
-		err = got_repo_cache_pack(pack, repo, path_packfile, packidx);
-		if (err)
-			goto done;
-	}
-	if ((*pack)->privsep_child == NULL) {
-		err = start_pack_privsep_child(*pack, packidx);
-		if (err)
-			goto done;
-	}
-
-	if (!(*pack)->child_has_delta_outfd) {
-		int outfd_child;
-		outfd_child = dup(delta_outfd);
-		if (outfd_child == -1) {
-			err = got_error_from_errno("dup");
-			goto done;
-		}
-		err = got_privsep_send_raw_delta_outfd(
-		    (*pack)->privsep_child->ibuf, outfd_child);
-		if (err)
-			goto done;
-		(*pack)->child_has_delta_outfd = 1;
-	}
-
-	err = got_privsep_send_delta_reuse_req((*pack)->privsep_child->ibuf);
-done:
-	free(path_packfile);
-	return err;
-}
-
 static const struct got_error *
 request_object(struct got_object **obj, struct got_object_id *id,
     struct got_repository *repo, int fd)
@@ -840,7 +716,7 @@ read_packed_commit_privsep(struct got_commit_object **commit,
 	if (pack->privsep_child)
 		return request_packed_commit(commit, pack, idx, id);
 
-	err = start_pack_privsep_child(pack, packidx);
+	err = got_pack_start_privsep_child(pack, packidx);
 	if (err)
 		return err;
 
@@ -1045,7 +921,7 @@ read_packed_tree_privsep(struct got_tree_object **tree,
 	if (pack->privsep_child)
 		return request_packed_tree(tree, pack, idx, id);
 
-	err = start_pack_privsep_child(pack, packidx);
+	err = got_pack_start_privsep_child(pack, packidx);
 	if (err)
 		return err;
 
@@ -1068,7 +944,7 @@ request_tree(struct got_tree_object **tree, struct got_repository *repo,
 	return got_privsep_recv_tree(tree, ibuf);
 }
 
-const struct got_error *
+static const struct got_error *
 read_tree_privsep(struct got_tree_object **tree, int obj_fd,
     struct got_object_id *id, struct got_repository *repo)
 {
@@ -1282,19 +1158,30 @@ got_tree_entry_get_symlink_target(char **link_target, struct got_tree_entry *te,
 {
 	const struct got_error *err = NULL;
 	struct got_blob_object *blob = NULL;
+	int fd = -1;
 
 	*link_target = NULL;
 
 	if (!got_object_tree_entry_is_symlink(te))
 		return got_error(GOT_ERR_TREE_ENTRY_TYPE);
 
+	fd = got_opentempfd();
+	if (fd == -1) {
+		err = got_error_from_errno("got_opentempfd");
+		goto done;
+	}
+
 	err = got_object_open_as_blob(&blob, repo,
-	    got_tree_entry_get_id(te), PATH_MAX);
+	    got_tree_entry_get_id(te), PATH_MAX, fd);
 	if (err)
-		return err;
+		goto done;
 
 	err = got_object_blob_read_to_str(link_target, blob);
-	got_object_blob_close(blob);
+done:
+	if (fd != -1 && close(fd) == -1 && err == NULL)
+		err = got_error_from_errno("close");
+	if (blob)
+		got_object_blob_close(blob);
 	if (err) {
 		free(*link_target);
 		*link_target = NULL;
@@ -1368,7 +1255,7 @@ read_packed_blob_privsep(uint8_t **outbuf, size_t *size, size_t *hdrlen,
 	const struct got_error *err = NULL;
 
 	if (pack->privsep_child == NULL) {
-		err = start_pack_privsep_child(pack, packidx);
+		err = got_pack_start_privsep_child(pack, packidx);
 		if (err)
 			return err;
 	}
@@ -1459,14 +1346,13 @@ read_blob_privsep(uint8_t **outbuf, size_t *size, size_t *hdrlen,
 
 static const struct got_error *
 open_blob(struct got_blob_object **blob, struct got_repository *repo,
-    struct got_object_id *id, size_t blocksize)
+    struct got_object_id *id, size_t blocksize, int outfd)
 {
 	const struct got_error *err = NULL;
 	struct got_packidx *packidx = NULL;
-	int idx;
+	int idx, dfd = -1;
 	char *path_packfile = NULL;
 	uint8_t *outbuf;
-	int outfd;
 	size_t size, hdrlen;
 	struct stat sb;
 
@@ -1474,16 +1360,21 @@ open_blob(struct got_blob_object **blob, struct got_repository *repo,
 	if (*blob == NULL)
 		return got_error_from_errno("calloc");
 
-	outfd = got_opentempfd();
-	if (outfd == -1)
-		return got_error_from_errno("got_opentempfd");
-
 	(*blob)->read_buf = malloc(blocksize);
 	if ((*blob)->read_buf == NULL) {
 		err = got_error_from_errno("malloc");
 		goto done;
 	}
 
+	if (ftruncate(outfd, 0L) == -1) {
+		err = got_error_from_errno("ftruncate");
+		goto done;
+	}
+	if (lseek(outfd, SEEK_SET, 0) == -1) {
+		err = got_error_from_errno("lseek");
+		goto done;
+	}
+
 	err = got_repo_search_packidx(&packidx, &idx, repo, id);
 	if (err == NULL) {
 		struct got_pack *pack = NULL;
@@ -1520,9 +1411,6 @@ open_blob(struct got_blob_object **blob, struct got_repository *repo,
 	}
 
 	if (outbuf) {
-		if (close(outfd) == -1 && err == NULL)
-			err = got_error_from_errno("close");
-		outfd = -1;
 		(*blob)->f = fmemopen(outbuf, size, "rb");
 		if ((*blob)->f == NULL) {
 			err = got_error_from_errno("fmemopen");
@@ -1541,11 +1429,17 @@ open_blob(struct got_blob_object **blob, struct got_repository *repo,
 			goto done;
 		}
 
-		(*blob)->f = fdopen(outfd, "rb");
+		dfd = dup(outfd);
+		if (dfd == -1) {
+			err = got_error_from_errno("dup");
+			goto done;
+		}
+
+		(*blob)->f = fdopen(dfd, "rb");
 		if ((*blob)->f == NULL) {
 			err = got_error_from_errno("fdopen");
-			close(outfd);
-			outfd = -1;
+			close(dfd);
+			dfd = -1;
 			goto done;
 		}
 	}
@@ -1560,25 +1454,25 @@ done:
 		if (*blob) {
 			got_object_blob_close(*blob);
 			*blob = NULL;
-		} else if (outfd != -1)
-			close(outfd);
+		}
 	}
 	return err;
 }
 
 const struct got_error *
 got_object_open_as_blob(struct got_blob_object **blob,
-    struct got_repository *repo, struct got_object_id *id,
-    size_t blocksize)
+    struct got_repository *repo, struct got_object_id *id, size_t blocksize,
+    int outfd)
 {
-	return open_blob(blob, repo, id, blocksize);
+	return open_blob(blob, repo, id, blocksize, outfd);
 }
 
 const struct got_error *
 got_object_blob_open(struct got_blob_object **blob,
-    struct got_repository *repo, struct got_object *obj, size_t blocksize)
+    struct got_repository *repo, struct got_object *obj, size_t blocksize,
+    int outfd)
 {
-	return open_blob(blob, repo, got_object_get_id(obj), blocksize);
+	return open_blob(blob, repo, got_object_get_id(obj), blocksize, outfd);
 }
 
 const struct got_error *
@@ -1744,7 +1638,7 @@ read_packed_tag_privsep(struct got_tag_object **tag,
 	if (pack->privsep_child)
 		return request_packed_tag(tag, pack, idx, id);
 
-	err = start_pack_privsep_child(pack, packidx);
+	err = got_pack_start_privsep_child(pack, packidx);
 	if (err)
 		return err;
 
@@ -2368,7 +2262,7 @@ got_traverse_packed_commits(struct got_object_id_queue *traversed_commits,
 	}
 
 	if (pack->privsep_child == NULL) {
-		err = start_pack_privsep_child(pack, packidx);
+		err = got_pack_start_privsep_child(pack, packidx);
 		if (err)
 			goto done;
 	}
@@ -2424,7 +2318,7 @@ got_object_enumerate(int *found_all_objects,
 	}
 
 	if (pack->privsep_child == NULL) {
-		err = start_pack_privsep_child(pack, packidx);
+		err = got_pack_start_privsep_child(pack, packidx);
 		if (err)
 			goto done;
 	}
