@@ -561,6 +561,28 @@ send_id(struct got_object_id *id, void *data, void *arg)
 	return NULL;
 }
 
+static const struct got_error *
+send_idset(struct imsgbuf *ibuf, struct got_object_idset *idset)
+{
+	const struct got_error *err;
+	struct send_id_arg sia;
+
+	memset(&sia, 0, sizeof(sia));
+	sia.ibuf = ibuf;
+	err = got_object_idset_for_each(idset, send_id, &sia);
+	if (err)
+		return err;
+
+	if (sia.nids > 0) {
+		err = got_privsep_send_object_idlist(ibuf, sia.ids, sia.nids);
+		if (err)
+			return err;
+	}
+
+	return got_privsep_send_object_idlist_done(ibuf);
+}
+
+
 static const struct got_error *
 recv_reused_delta(struct got_imsg_reused_delta *delta,
     struct got_object_idset *idset, struct got_pack_metavec *v)
@@ -593,6 +615,60 @@ recv_reused_delta(struct got_imsg_reused_delta *delta,
 	return add_meta(m, v);
 }
 
+static const struct got_error *
+cache_pack_for_packidx(struct got_pack **pack, struct got_packidx *packidx,
+    struct got_repository *repo)
+{
+	const struct got_error *err;
+	char *path_packfile = NULL;
+
+	err = got_packidx_get_packfile_path(&path_packfile,
+	    packidx->path_packidx);
+	if (err)
+		return err;
+
+	*pack = got_repo_get_cached_pack(repo, path_packfile);
+	if (*pack == NULL) {
+		err = got_repo_cache_pack(pack, repo, path_packfile, packidx);
+		if (err)
+			goto done;
+	}
+	if ((*pack)->privsep_child == NULL) {
+		err = got_pack_start_privsep_child(*pack, packidx);
+		if (err)
+			goto done;
+	}
+done:
+	free(path_packfile);
+	return err;
+}
+
+static const struct got_error *
+prepare_delta_reuse(struct got_pack *pack, struct got_packidx *packidx,
+    int delta_outfd, struct got_repository *repo)
+{
+	const struct got_error *err = NULL;
+
+	if (!pack->child_has_delta_outfd) {
+		int outfd_child;
+		outfd_child = dup(delta_outfd);
+		if (outfd_child == -1) {
+			err = got_error_from_errno("dup");
+			goto done;
+		}
+		err = got_privsep_send_raw_delta_outfd(
+		    pack->privsep_child->ibuf, outfd_child);
+		if (err)
+			goto done;
+		pack->child_has_delta_outfd = 1;
+	}
+
+	err = got_privsep_send_delta_reuse_req(pack->privsep_child->ibuf);
+done:
+	return err;
+}
+
+
 static const struct got_error *
 search_deltas(struct got_pack_metavec *v, struct got_object_idset *idset,
     int delta_cache_fd, int ncolored, int nfound, int ntrees, int ncommits,
@@ -603,7 +679,6 @@ search_deltas(struct got_pack_metavec *v, struct got_object_idset *idset,
 	const struct got_error *err = NULL;
 	struct got_packidx *packidx;
 	struct got_pack *pack;
-	struct send_id_arg sia;
 	struct got_imsg_reused_delta deltas[GOT_IMSG_REUSED_DELTAS_MAX_NDELTAS];
 	size_t ndeltas, i;
 
@@ -614,23 +689,15 @@ search_deltas(struct got_pack_metavec *v, struct got_object_idset *idset,
 	if (packidx == NULL)
 		return NULL;
 
-	err = got_object_prepare_delta_reuse(&pack, packidx,
-	    delta_cache_fd, repo);
+	err = cache_pack_for_packidx(&pack, packidx, repo);
 	if (err)
 		return err;
 
-	memset(&sia, 0, sizeof(sia));
-	sia.ibuf = pack->privsep_child->ibuf;
-	err = got_object_idset_for_each(idset, send_id, &sia);
+	err = prepare_delta_reuse(pack, packidx, delta_cache_fd, repo);
 	if (err)
 		return err;
-	if (sia.nids > 0) {
-		err = got_privsep_send_object_idlist(pack->privsep_child->ibuf,
-		    sia.ids, sia.nids);
-		if (err)
-			return err;
-	}
-	err = got_privsep_send_object_idlist_done(pack->privsep_child->ibuf);
+
+	err = send_idset(pack->privsep_child->ibuf, idset);
 	if (err)
 		return err;
 
@@ -1172,30 +1239,23 @@ done:
 enum findtwixt_color {
 	COLOR_KEEP = 0,
 	COLOR_DROP,
-	COLOR_BLANK,
-	COLOR_SKIP,
-};
-
-static const int findtwixt_colors[] = {
-	COLOR_KEEP,
-	COLOR_DROP,
-	COLOR_BLANK,
 	COLOR_SKIP,
+	COLOR_MAX,
 };
 
 static const struct got_error *
-paint_commit(struct got_object_qid *qid, int color)
+paint_commit(struct got_object_qid *qid, intptr_t color)
 {
-	if (color < 0 || color >= nitems(findtwixt_colors))
+	if (color < 0 || color >= COLOR_MAX)
 		return got_error(GOT_ERR_RANGE);
 
-	qid->data = (void *)&findtwixt_colors[color];
+	qid->data = (void *)color;
 	return NULL;
 }
 
 static const struct got_error *
 queue_commit_id(struct got_object_id_queue *ids, struct got_object_id *id,
-    int color, struct got_repository *repo)
+    intptr_t color, struct got_repository *repo)
 {
 	const struct got_error *err;
 	struct got_object_qid *qid;
@@ -1232,7 +1292,7 @@ append_id(struct got_object_id *id, void *data, void *arg)
 }
 
 static const struct got_error *
-queue_commit_or_tag_id(struct got_object_id *id, int color,
+queue_commit_or_tag_id(struct got_object_id *id, intptr_t color,
     struct got_object_id_queue *ids, struct got_repository *repo)
 {
 	const struct got_error *err;
@@ -1262,6 +1322,196 @@ done:
 	return err;
 }
 
+struct recv_painted_commit_arg {
+	int *ncolored;
+	int *nqueued;
+	int *nskip;
+	struct got_object_id_queue *ids;
+	struct got_object_idset *keep;
+	struct got_object_idset *drop;
+	struct got_object_idset *skip;
+	got_pack_progress_cb progress_cb;
+	void *progress_arg;
+	struct got_ratelimit *rl;
+	got_cancel_cb cancel_cb;
+	void *cancel_arg;
+};
+
+static const struct got_error *
+recv_painted_commit(void *arg, struct got_object_id *id, intptr_t color)
+{
+	const struct got_error *err = NULL;
+	struct recv_painted_commit_arg *a = arg;
+	struct got_object_qid *qid, *tmp;
+
+	if (a->cancel_cb) {
+		err = a->cancel_cb(a->cancel_arg);
+		if (err)
+			return err;
+	}
+
+	switch (color) {
+	case COLOR_KEEP:
+		err = got_object_idset_add(a->keep, id, NULL);
+		if (err)
+			return err;
+		(*a->ncolored)++;
+		break;
+	case COLOR_DROP:
+		err = got_object_idset_add(a->drop, id, NULL);
+		if (err)
+			return err;
+		(*a->ncolored)++;
+		break;
+	case COLOR_SKIP:
+		err = got_object_idset_add(a->skip, id, NULL);
+		if (err)
+			return err;
+		break;
+	default:
+		/* should not happen */
+		return got_error_fmt(GOT_ERR_NOT_IMPL,
+		    "%s invalid commit color %d", __func__, color);
+	}
+
+	STAILQ_FOREACH_SAFE(qid, a->ids, entry, tmp) {
+		if (got_object_id_cmp(&qid->id, id) != 0)
+			continue;
+		STAILQ_REMOVE(a->ids, qid, got_object_qid, entry);
+		color = (intptr_t)qid->data;
+		got_object_qid_free(qid);
+		(*a->nqueued)--;
+		if (color == COLOR_SKIP)
+			(*a->nskip)--;
+		break;
+	}
+
+	return report_progress(a->progress_cb, a->progress_arg, a->rl,
+	    *a->ncolored, 0, 0, 0L, 0, 0, 0, 0);
+}
+
+static const struct got_error *
+paint_packed_commits(struct got_pack *pack, struct got_object_id *id,
+    int idx, intptr_t color, int *ncolored, int *nqueued, int *nskip,
+    struct got_object_id_queue *ids,
+    struct got_object_idset *keep, struct got_object_idset *drop,
+    struct got_object_idset *skip, struct got_repository *repo,
+    got_pack_progress_cb progress_cb, void *progress_arg,
+    struct got_ratelimit *rl, got_cancel_cb cancel_cb, void *cancel_arg)
+{
+	const struct got_error *err = NULL;
+	struct got_object_id_queue next_ids;
+	struct got_object_qid *qid, *tmp;
+	struct recv_painted_commit_arg arg;
+
+	STAILQ_INIT(&next_ids);
+
+	err = got_privsep_send_painting_request(pack->privsep_child->ibuf,
+	    idx, id, color);
+	if (err)
+		return err;
+
+	arg.ncolored = ncolored;
+	arg.nqueued = nqueued;
+	arg.nskip = nskip;
+	arg.ids = ids;
+	arg.keep = keep;
+	arg.drop = drop;
+	arg.skip = skip;
+	arg.progress_cb = progress_cb;
+	arg.progress_arg = progress_arg;
+	arg.rl = rl;
+	arg.cancel_cb = cancel_cb;
+	arg.cancel_arg = cancel_arg;
+	err = got_privsep_recv_painted_commits(&next_ids,
+	    recv_painted_commit, &arg, pack->privsep_child->ibuf);
+	if (err)
+		return err;
+
+	STAILQ_FOREACH_SAFE(qid, &next_ids, entry, tmp) {
+		struct got_object_qid *old_id;
+		intptr_t qcolor, ocolor;
+		STAILQ_FOREACH(old_id, ids, entry) {
+			if (got_object_id_cmp(&qid->id, &old_id->id))
+				continue;
+			qcolor = (intptr_t)qid->data;
+			ocolor = (intptr_t)old_id->data;
+			STAILQ_REMOVE(&next_ids, qid, got_object_qid, entry);
+			got_object_qid_free(qid);
+			qid = NULL;
+			if (qcolor != ocolor) {
+				paint_commit(old_id, qcolor);
+				if (ocolor == COLOR_SKIP)
+					(*nskip)--;
+				else if (qcolor == COLOR_SKIP)
+					(*nskip)++;
+			}
+			break;
+		}
+	}
+	while (!STAILQ_EMPTY(&next_ids)) {
+		qid = STAILQ_FIRST(&next_ids);
+		STAILQ_REMOVE_HEAD(&next_ids, entry);
+		paint_commit(qid, color);
+		STAILQ_INSERT_TAIL(ids, qid, entry);
+		(*nqueued)++;
+		if (color == COLOR_SKIP)
+			(*nskip)++;
+	}
+
+	return err;
+}
+
+static const struct got_error *
+find_pack_for_commit_painting(struct got_packidx **best_packidx,
+    struct got_object_id_queue *ids, int nids, struct got_repository *repo)
+{
+	const struct got_error *err = NULL;
+	struct got_pathlist_entry *pe;
+	const char *best_packidx_path = NULL;
+	int nobj_max = 0;
+	int ncommits_max = 0;
+
+	*best_packidx = NULL;
+
+	/*
+	 * Find the largest pack which contains at least some of the
+	 * commits we are interested in.
+	 */
+	TAILQ_FOREACH(pe, &repo->packidx_paths, entry) {
+		const char *path_packidx = pe->path;
+		struct got_packidx *packidx;
+		int nobj, idx, ncommits = 0;
+		struct got_object_qid *qid;
+
+		err = got_repo_get_packidx(&packidx, path_packidx, repo);
+		if (err)
+			break;
+
+		nobj = be32toh(packidx->hdr.fanout_table[0xff]);
+		if (nobj <= nobj_max)
+			continue;
+
+		STAILQ_FOREACH(qid, ids, entry) {
+			idx = got_packidx_get_object_idx(packidx, &qid->id);
+			if (idx != -1)
+				ncommits++;
+		}
+		if (ncommits > ncommits_max) {
+			best_packidx_path = path_packidx;
+			nobj_max = nobj;
+			ncommits_max = ncommits;
+		}
+	}
+
+	if (best_packidx_path && err == NULL) {
+		err = got_repo_get_packidx(best_packidx, best_packidx_path,
+		    repo);
+	}
+
+	return err;
+}
+
 static const struct got_error *
 paint_commits(int *ncolored, struct got_object_id_queue *ids, int nids,
     struct got_object_idset *keep, struct got_object_idset *drop,
@@ -1271,12 +1521,15 @@ paint_commits(int *ncolored, struct got_object_id_queue *ids, int nids,
 {
 	const struct got_error *err = NULL;
 	struct got_commit_object *commit = NULL;
+	struct got_packidx *packidx = NULL;
+	struct got_pack *pack = NULL;
 	const struct got_object_id_queue *parents;
-	struct got_object_qid *qid;
+	struct got_object_qid *qid = NULL;
 	int nqueued = nids, nskip = 0;
+	int idx;
 
 	while (!STAILQ_EMPTY(ids) && nskip != nqueued) {
-		int color;
+		intptr_t color;
 
 		if (cancel_cb) {
 			err = cancel_cb(cancel_arg);
@@ -1287,21 +1540,49 @@ paint_commits(int *ncolored, struct got_object_id_queue *ids, int nids,
 		qid = STAILQ_FIRST(ids);
 		STAILQ_REMOVE_HEAD(ids, entry);
 		nqueued--;
-		color = *((int *)qid->data);
+		color = (intptr_t)qid->data;
 		if (color == COLOR_SKIP)
 			nskip--;
 
 		if (got_object_idset_contains(skip, &qid->id)) {
 			got_object_qid_free(qid);
+			qid = NULL;
 			continue;
 		}
+		if (color == COLOR_KEEP &&
+		    got_object_idset_contains(keep, &qid->id)) {
+			got_object_qid_free(qid);
+			qid = NULL;
+			continue;
+		}
+		if (color == COLOR_DROP &&
+		    got_object_idset_contains(drop, &qid->id)) {
+			got_object_qid_free(qid);
+			qid = NULL;
+			continue;
+		}
+
+		/* Pinned pack may have moved to different cache slot. */
+		pack = got_repo_get_pinned_pack(repo);
+
+		if (packidx && pack) {
+			idx = got_packidx_get_object_idx(packidx, &qid->id);
+			if (idx != -1) {
+				err = paint_packed_commits(pack, &qid->id,
+				    idx, color, ncolored, &nqueued, &nskip,
+				    ids, keep, drop, skip, repo,
+				    progress_cb, progress_arg, rl,
+				    cancel_cb, cancel_arg);
+				if (err)
+					break;
+				got_object_qid_free(qid);
+				qid = NULL;
+				continue;
+			}
+		}
 
 		switch (color) {
 		case COLOR_KEEP:
-			if (got_object_idset_contains(keep, &qid->id)) {
-				got_object_qid_free(qid);
-				continue;
-			}
 			if (got_object_idset_contains(drop, &qid->id)) {
 				err = paint_commit(qid, COLOR_SKIP);
 				if (err)
@@ -1313,10 +1594,6 @@ paint_commits(int *ncolored, struct got_object_id_queue *ids, int nids,
 				goto done;
 			break;
 		case COLOR_DROP:
-			if (got_object_idset_contains(drop, &qid->id)) {
-				got_object_qid_free(qid);
-				continue;
-			}
 			if (got_object_idset_contains(keep, &qid->id)) {
 				err = paint_commit(qid, COLOR_SKIP);
 				if (err)
@@ -1347,7 +1624,6 @@ paint_commits(int *ncolored, struct got_object_id_queue *ids, int nids,
 		if (err)
 			break;
 
-
 		err = got_object_open_as_commit(&commit, repo, &qid->id);
 		if (err)
 			break;
@@ -1355,10 +1631,10 @@ paint_commits(int *ncolored, struct got_object_id_queue *ids, int nids,
 		parents = got_object_commit_get_parent_ids(commit);
 		if (parents) {
 			struct got_object_qid *pid;
-			color = *((int *)qid->data);
+			color = (intptr_t)qid->data;
 			STAILQ_FOREACH(pid, parents, entry) {
-				err = queue_commit_id(ids, &pid->id, color,
-				    repo);
+				err = queue_commit_id(ids, &pid->id,
+				    color, repo);
 				if (err)
 					break;
 				nqueued++;
@@ -1367,13 +1643,56 @@ paint_commits(int *ncolored, struct got_object_id_queue *ids, int nids,
 			}
 		}
 
+		if (pack == NULL && (commit->flags & GOT_COMMIT_FLAG_PACKED)) {
+			if (packidx == NULL) {
+				err = find_pack_for_commit_painting(&packidx,
+				    ids, nqueued, repo);
+				if (err)
+					goto done;
+			}
+			if (packidx != NULL) {
+				err = cache_pack_for_packidx(&pack, packidx,
+				    repo);
+				if (err)
+					goto done;
+				err = got_privsep_init_commit_painting(
+				    pack->privsep_child->ibuf);
+				if (err)
+					goto done;
+				err = send_idset(pack->privsep_child->ibuf,
+				    keep);
+				if (err)
+					goto done;
+				err = send_idset(pack->privsep_child->ibuf, drop);
+				if (err)
+					goto done;
+				err = send_idset(pack->privsep_child->ibuf, skip);
+				if (err)
+					goto done;
+				err = got_repo_pin_pack(repo, packidx, pack);
+				if (err)
+					goto done;
+			}
+		}
+
 		got_object_commit_close(commit);
 		commit = NULL;
+
 		got_object_qid_free(qid);
+		qid = NULL;
 	}
 done:
+	if (pack) {
+		const struct got_error *pack_err;
+		pack_err = got_privsep_send_painting_commits_done(
+		    pack->privsep_child->ibuf);
+		if (err == NULL)
+			err = pack_err;
+	}
 	if (commit)
 		got_object_commit_close(commit);
+	got_object_qid_free(qid);
+	got_repo_unpin_pack(repo);
 	return err;
 }
 
@@ -1783,8 +2102,8 @@ done:
 	return err;
 }
 
-const struct got_error *
-hwrite(FILE *f, void *buf, off_t len, SHA1_CTX *ctx)
+static const struct got_error *
+hwrite(FILE *f, const void *buf, off_t len, SHA1_CTX *ctx)
 {
 	size_t n;
 
@@ -1795,7 +2114,7 @@ hwrite(FILE *f, void *buf, off_t len, SHA1_CTX *ctx)
 	return NULL;
 }
 
-const struct got_error *
+static const struct got_error *
 hcopy(FILE *fsrc, FILE *fdst, off_t len, SHA1_CTX *ctx)
 {
 	unsigned char buf[65536];
@@ -1817,7 +2136,7 @@ hcopy(FILE *fsrc, FILE *fdst, off_t len, SHA1_CTX *ctx)
 	return NULL;
 }
 
-const struct got_error *
+static const struct got_error *
 hcopy_mmap(uint8_t *src, off_t src_offset, size_t src_size,
     FILE *fdst, off_t len, SHA1_CTX *ctx)
 {
