@@ -25,6 +25,7 @@
 #include <sys/stat.h>
 #include <sys/uio.h>
 
+#include <ctype.h>
 #include <errno.h>
 #include <limits.h>
 #include <stdint.h>
@@ -42,6 +43,7 @@
 #include "got_repository.h"
 #include "got_opentemp.h"
 #include "got_patch.h"
+#include "got_diff.h"
 
 #include "got_lib_delta.h"
 #include "got_lib_diff.h"
@@ -54,6 +56,7 @@
 struct got_patch_hunk {
 	STAILQ_ENTRY(got_patch_hunk) entries;
 	const struct got_error *err;
+	int	ws_mangled;
 	int	offset;
 	int	old_nonl;
 	int	new_nonl;
@@ -189,10 +192,11 @@ recv_patch(struct imsgbuf *ibuf, int *done, struct got_patch *p, int strip)
 		goto done;
 	}
 
-	if (*patch.cid != '\0' && *patch.blob != '\0') {
+	if (*patch.cid != '\0')
 		strlcpy(p->cid, patch.cid, sizeof(p->cid));
+
+	if (*patch.blob != '\0')
 		strlcpy(p->blob, patch.blob, sizeof(p->blob));
-	}
 
 	/* automatically set strip=1 for git-style diffs */
 	if (strip == -1 && patch.git &&
@@ -396,6 +400,30 @@ locate_hunk(FILE *orig, struct got_patch_hunk *h, off_t *pos, int *lineno)
 	return err;
 }
 
+static int
+linecmp(const char *a, const char *b, int *mangled)
+{
+	int c;
+
+	*mangled = 0;
+	c = strcmp(a, b);
+	if (c == 0)
+		return c;
+
+	*mangled = 1;
+	for (;;) {
+		while (*a == '\t' || *a == ' ' || *a == '\f')
+			a++;
+		while (*b == '\t' || *b == ' ' || *b == '\f')
+			b++;
+		if (*a == '\0' || *a != *b)
+			break;
+		a++, b++;
+	}
+
+	return *a - *b;
+}
+
 static const struct got_error *
 test_hunk(FILE *orig, struct got_patch_hunk *h)
 {
@@ -403,6 +431,7 @@ test_hunk(FILE *orig, struct got_patch_hunk *h)
 	char *line = NULL;
 	size_t linesize = 0, i = 0;
 	ssize_t linelen;
+	int mangled;
 
 	for (i = 0; i < h->len; ++i) {
 		switch (*h->lines[i]) {
@@ -421,10 +450,70 @@ test_hunk(FILE *orig, struct got_patch_hunk *h)
 			}
 			if (line[linelen - 1] == '\n')
 				line[linelen - 1] = '\0';
-			if (strcmp(h->lines[i] + 1, line)) {
+			if (linecmp(h->lines[i] + 1, line, &mangled)) {
 				err = got_error(GOT_ERR_HUNK_FAILED);
 				goto done;
 			}
+			if (mangled)
+				h->ws_mangled = 1;
+			break;
+		}
+	}
+
+done:
+	free(line);
+	return err;
+}
+
+static const struct got_error *
+apply_hunk(FILE *orig, FILE *tmp, struct got_patch_hunk *h, int *lineno,
+    off_t from)
+{
+	const struct got_error *err = NULL;
+	const char *t;
+	size_t linesize = 0, i, new = 0;
+	char *line = NULL;
+	char mode;
+	ssize_t linelen;
+
+	if (orig != NULL && fseeko(orig, from, SEEK_SET) == -1)
+		return got_error_from_errno("fseeko");
+
+	for (i = 0; i < h->len; ++i) {
+		switch (mode = *h->lines[i]) {
+		case '-':
+		case ' ':
+			(*lineno)++;
+			if (orig != NULL) {
+				linelen = getline(&line, &linesize, orig);
+				if (linelen == -1) {
+					err = got_error_from_errno("getline");
+					goto done;
+				}
+				if (line[linelen - 1] == '\n')
+					line[linelen - 1] = '\0';
+				t = line;
+			} else
+				t = h->lines[i] + 1;
+			if (mode == '-')
+				continue;
+			if (fprintf(tmp, "%s\n", t) < 0) {
+				err = got_error_from_errno("fprintf");
+				goto done;
+			}
+			break;
+		case '+':
+			new++;
+			if (fprintf(tmp, "%s", h->lines[i] + 1) < 0) {
+				err = got_error_from_errno("fprintf");
+				goto done;
+			}
+			if (new != h->new_lines || !h->new_nonl) {
+				if (fprintf(tmp, "\n") < 0) {
+					err = got_error_from_errno("fprintf");
+					goto done;
+				}
+			}
 			break;
 		}
 	}
@@ -435,36 +524,7 @@ done:
 }
 
 static const struct got_error *
-apply_hunk(FILE *tmp, struct got_patch_hunk *h, int *lineno)
-{
-	size_t i, new = 0;
-
-	for (i = 0; i < h->len; ++i) {
-		switch (*h->lines[i]) {
-		case ' ':
-			if (fprintf(tmp, "%s\n", h->lines[i] + 1) < 0)
-				return got_error_from_errno("fprintf");
-			/* fallthrough */
-		case '-':
-			(*lineno)++;
-			break;
-		case '+':
-			new++;
-			if (fprintf(tmp, "%s", h->lines[i] + 1) < 0)
-				return got_error_from_errno("fprintf");
-			if (new != h->new_lines || !h->new_nonl) {
-				if (fprintf(tmp, "\n") < 0)
-					return got_error_from_errno(
-					    "fprintf");
-			}
-			break;
-		}
-	}
-	return NULL;
-}
-
-static const struct got_error *
-patch_file(struct got_patch *p, FILE *orig, FILE *tmp, mode_t *mode)
+patch_file(struct got_patch *p, FILE *orig, FILE *tmp)
 {
 	const struct got_error *err = NULL;
 	struct got_patch_hunk *h;
@@ -479,12 +539,11 @@ patch_file(struct got_patch *p, FILE *orig, FILE *tmp, mode_t *mode)
 		h = STAILQ_FIRST(&p->head);
 		if (h == NULL || STAILQ_NEXT(h, entries) != NULL)
 			return got_error(GOT_ERR_PATCH_MALFORMED);
-		return apply_hunk(tmp, h, &lineno);
+		return apply_hunk(orig, tmp, h, &lineno, 0);
 	}
 
 	if (fstat(fileno(orig), &sb) == -1)
 		return got_error_from_errno("fstat");
-	*mode = sb.st_mode;
 
 	copypos = 0;
 	STAILQ_FOREACH(h, &p->head, entries) {
@@ -519,7 +578,7 @@ patch_file(struct got_patch *p, FILE *orig, FILE *tmp, mode_t *mode)
 		if (lineno + 1 != h->old_from)
 			h->offset = lineno + 1 - h->old_from;
 
-		err = apply_hunk(tmp, h, &lineno);
+		err = apply_hunk(orig, tmp, h, &lineno, pos);
 		if (err != NULL)
 			return err;
 
@@ -546,17 +605,17 @@ report_progress(struct patch_args *pa, const char *old, const char *new,
 	struct got_patch_hunk *h;
 
 	err = pa->progress_cb(pa->progress_arg, old, new, status,
-	    orig_error, 0, 0, 0, 0, 0, NULL);
+	    orig_error, 0, 0, 0, 0, 0, 0, NULL);
 	if (err)
 		return err;
 
 	STAILQ_FOREACH(h, pa->head, entries) {
-		if (h->offset == 0 && h->err == NULL)
+		if (h->offset == 0 && !h->ws_mangled && h->err == NULL)
 			continue;
 
 		err = pa->progress_cb(pa->progress_arg, old, new, 0, NULL,
 		    h->old_from, h->old_lines, h->new_from, h->new_lines,
-		    h->offset, h->err);
+		    h->offset, h->ws_mangled, h->err);
 		if (err)
 			return err;
 	}
@@ -583,15 +642,32 @@ open_blob(char **path, FILE **fp, const char *blobid,
 {
 	const struct got_error *err = NULL;
 	struct got_blob_object *blob = NULL;
-	struct got_object_id id;
+	struct got_object_id id, *idptr, *matched_id = NULL;
+	int fd = -1;
 
 	*fp = NULL;
 	*path = NULL;
 
-	if (!got_parse_sha1_digest(id.sha1, blobid))
-		return got_error(GOT_ERR_BAD_OBJ_ID_STR);
+	if (strlen(blobid) != SHA1_DIGEST_STRING_LENGTH - 1) {
+		err = got_repo_match_object_id(&matched_id, NULL, blobid,
+		    GOT_OBJ_TYPE_BLOB, NULL /* do not resolve tags */,
+		    repo);
+		if (err)
+			return err;
+		idptr = matched_id;
+	} else {
+		if (!got_parse_sha1_digest(id.sha1, blobid))
+			return got_error(GOT_ERR_BAD_OBJ_ID_STR);
+		idptr = &id;
+	}
 
-	err = got_object_open_as_blob(&blob, repo, &id, 8192);
+	fd = got_opentempfd();
+	if (fd == -1) {
+		err = got_error_from_errno("got_opentempfd");
+		goto done;
+	}
+
+	err = got_object_open_as_blob(&blob, repo, idptr, 8192, fd);
 	if (err)
 		goto done;
 
@@ -604,8 +680,12 @@ open_blob(char **path, FILE **fp, const char *blobid,
 		goto done;
 
 done:
+	if (fd != -1 && close(fd) == -1 && err == NULL)
+		err = got_error_from_errno("close");
 	if (blob)
 		got_object_blob_close(blob);
+	if (matched_id != NULL)
+		free(matched_id);
 	if (err) {
 		if (*fp != NULL)
 			fclose(*fp);
@@ -625,6 +705,7 @@ apply_patch(int *overlapcnt, struct got_worktree *worktree,
     struct patch_args *pa, got_cancel_cb cancel_cb, void *cancel_arg)
 {
 	const struct got_error *err = NULL;
+	struct stat sb;
 	int do_merge = 0, file_renamed = 0;
 	char *oldlabel = NULL, *newlabel = NULL, *anclabel = NULL;
 	char *oldpath = NULL, *newpath = NULL;
@@ -644,7 +725,8 @@ apply_patch(int *overlapcnt, struct got_worktree *worktree,
 		 * ignore failures to open this blob, we might have
 		 * parsed gibberish.
 		 */
-		if (err && !(err->code == GOT_ERR_ERRNO && errno == ENOENT))
+		if (err && !(err->code == GOT_ERR_ERRNO && errno == ENOENT) &&
+		    err->code != GOT_ERR_NO_OBJ)
 			return err;
 		else if (err == NULL)
 			do_merge = 1;
@@ -671,9 +753,16 @@ apply_patch(int *overlapcnt, struct got_worktree *worktree,
 		goto done;
 	}
 
-	if (p->old != NULL && (oldfile = fopen(oldpath, "r")) == NULL) {
-		err = got_error_from_errno2("open", oldpath);
-		goto done;
+	if (p->old != NULL) {
+		if ((oldfile = fopen(oldpath, "r")) == NULL) {
+			err = got_error_from_errno2("open", oldpath);
+			goto done;
+		}
+		if (fstat(fileno(oldfile), &sb) == -1) {
+			err = got_error_from_errno2("fstat", oldpath);
+			goto done;
+		}
+		mode = sb.st_mode;
 	}
 
 	err = got_opentemp_named(&tmppath, &tmpfile, template);
@@ -681,11 +770,13 @@ apply_patch(int *overlapcnt, struct got_worktree *worktree,
 		goto done;
 	outpath = tmppath;
 	outfd = fileno(tmpfile);
-	err = patch_file(p, afile != NULL ? afile : oldfile, tmpfile, &mode);
+	err = patch_file(p, afile != NULL ? afile : oldfile, tmpfile);
 	if (err)
 		goto done;
 
 	if (do_merge) {
+		const char *type, *id;
+
 		if (fseeko(afile, 0, SEEK_SET) == -1 ||
 		    fseeko(oldfile, 0, SEEK_SET) == -1 ||
 		    fseeko(tmpfile, 0, SEEK_SET) == -1) {
@@ -705,7 +796,15 @@ apply_patch(int *overlapcnt, struct got_worktree *worktree,
 			goto done;
 		}
 
-		if (asprintf(&anclabel, "commit %s", p->cid) == -1) {
+		if (*p->cid != '\0') {
+			type = "commit";
+			id = p->cid;
+		} else {
+			type = "blob";
+			id = p->blob;
+		}
+
+		if (asprintf(&anclabel, "%s %s", type, id) == -1) {
 			err = got_error_from_errno("asprintf");
 			anclabel = NULL;
 			goto done;
